# -*- coding: utf-8 -*-
"""Stack tests."""

import numpy
import pytest
import scopesim
from micado.config.Environment import Env
from micado.dataitems.Calibrated import Calibrated
from micado.dataitems.Detrend import Detrend
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.Science import Science
from micado.dataitems.ScienceExtension import ScienceExtension
from micado.dataitems.Stack import Stack
from micado.dataitems.DepersistedScience import DepersistedScience
from micado.hardware.DarkCurrent import DarkCurrent
from micado.hardware.Detector import Detector
from micado.hardware.DetectorArray import DetectorArray
from micado.hardware.QuantumEfficiencyCurve import QuantumEfficiencyCurve
from micado.hardware.ShotNoise import ShotNoise
from micado.hardware.SummedExposure import SummedExposure
from micado.main.ObservingBlock import ObservingBlock
from micado.recipes.McdImgCalibrate import McdImgCalibrate
from micado.recipes.McdImgDetrend import McdImgDetrend
from micado.recipes.McdImgDetrendFinal import McdImgDetrendFinal
from micado.recipes.McdImgStack import McdImgStack
from micado.templates.MicadoImgObs import MicadoImgObs
from micado.test.functional.masterdark_test import mock_depersist
from micado.test.functional.science_test import get_filt_inst_ro, get_atmo_telescope


@pytest.mark.xfail(
    reason="Will fail because simulating a cluster will download from http://svo2.cab.inta-csic.es/theory/fps3/fps.php which is down. TODO: Put this back when we can simulate a cluster again, when is back up."
)
def test_stack():
    """Test stacking of images."""
    Env["use_python_logging"] = True

    scopesim.server.download_package("telescopes/ELT")
    scopesim.server.download_package("instruments/MICADO")
    scopesim.server.download_package("instruments/MAORY")
    scopesim.server.download_package("locations/Armazones")

    filt1, filt2, inst, ro = get_filt_inst_ro()
    atmo, telescope = get_atmo_telescope()

    myexposureaction = SummedExposure()

    # Full detector: hw = 30.72
    hw = 0.24 * 48

    mydetector = Detector(xhw=hw, yhw=hw)
    mydetectorlist = [mydetector]
    mydetectorarray = DetectorArray(
        name="testdetector",
        detector_array_list=mydetectorlist,
        exposure_action=myexposureaction,
        dark_current=DarkCurrent(value=2.0),
        shot_noise=ShotNoise(state=True),
        qecurve=QuantumEfficiencyCurve(realfilename="QE_detector_H2RG.dat"),
        # Linearity is disabled because this causes too many saturated pixels:
        #     linearity=LinearityCurve(realfilename="FPA_linearity.dat"),
    )

    # A template
    science_template = MicadoImgObs(det__dit=1)

    # An observingblock
    science_ob = ObservingBlock(
        name="MyScienceTest",
        instrument="MICADO",
        target="cluster",
        templates=[science_template],
    )

    myextensions = [
        ScienceExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    science1 = Science(
        detector=mydetectorarray,
        instrument=inst,
        relay_optics=ro,
        atmosphere=atmo,
        telescope=telescope,
        template=science_template,
        observing_block=science_ob,
        extensions=myextensions,
        tpl__expno=0,
    )

    science1.make_simulation()

    myextensions = [
        ScienceExtension(detector=det, BITPIX=-64, NAXIS=2) for det in mydetectorlist
    ]

    science3 = Science(
        detector=mydetectorarray,
        instrument=inst,
        relay_optics=ro,
        atmosphere=atmo,
        telescope=telescope,
        template=science_template,
        observing_block=science_ob,
        extensions=myextensions,
        tpl__expno=2,
    )

    science3.make_simulation()

    [depersist1, depersist3] = [
        mock_depersist(DepersistedScience, science) for science in [science1, science3]
    ]

    detrend1 = Detrend(
        raw=depersist1,
        recipe=McdImgDetrend(),
    )
    detrend1.make_proto()

    detrend3 = Detrend(
        raw=depersist3,
        recipe=McdImgDetrend(),
    )
    detrend3.make_proto()

    detrendfinal1 = DetrendFinal(
        detrend=detrend1,
        recipe=McdImgDetrendFinal(),
    )
    detrendfinal1.make_proto()

    detrendfinal3 = DetrendFinal(
        detrend=detrend3,
        recipe=McdImgDetrendFinal(),
    )
    detrendfinal3.make_proto()

    calibrateds = [
        Calibrated(
            detrend=detrend,
            recipe=McdImgCalibrate(),
        )
        for detrend in [detrendfinal1, detrendfinal3]
    ]

    for calib in calibrateds:
        calib.make_proto()

    stack = Stack(
        detrends=calibrateds,
        recipe=McdImgStack(),
    )

    stack.make_proto()

    data = stack.extensions[0].image
    print(data.shape)
    dmin, dmax, dmean, dmed, dstd = (
        data.min(),
        data.max(),
        data.mean(),
        numpy.median(data),
        data.std(),
    )
    print(dmin, dmax, dmean, dmed, dstd)
    # Typical: 16.0 10776465.5 455.00663871033476 85.0 22256.13457409676

    assert data.shape == (1786, 2286)

    # Some sanity checks for the values
    assert 1 < dmin < 32
    assert 1e6 < dmax < 1e9
    assert 1e2 < dmean < 1e3
    assert 10 < dmed < 200
    assert 1e3 < dstd < 1e6
