"""Generate LaTeX files from the Python classes."""
import itertools
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import List, Type, Tuple
import re

from micado.main.commonwrapper import classcache
from micado.dataitems.AstromRef import AstromRef
from micado.dataitems.Astromfield import Astromfield
from micado.dataitems.Background import Background
from micado.dataitems.CalibHCI import CalibHCI
from micado.dataitems.CubeHCI import CubeHCI
from micado.dataitems.DepersistedDark import DepersistedDark
from micado.dataitems.DepersistedNonlin import DepersistedNonlin
from micado.dataitems.DepersistedOpt import DepersistedOpt
from micado.dataitems.DetrendBase import DetrendBase
from micado.dataitems.DetrendFinal import DetrendFinal
from micado.dataitems.DetrendFinalBase import DetrendFinalBase

# The import of DpsicsStaroffsetOpt is only to ensure order in
# 'Used in recipes', e.g. for DerivePositionStar.tex.
# TODO: Fix this such that import ordering is irrelevant.
from micado.dataitems.DetrendFinalHCIBase import DetrendFinalHCIBase
from micado.dataitems.DetrendFinalPupil import DetrendFinalPupil
from micado.dataitems.DetrendFinalSiaiBase import DetrendFinalSiaiBase
from micado.dataitems.DetrendFinalWaffle import DetrendFinalWaffle
from micado.dataitems.DetrendHCIBase import DetrendHCIBase
from micado.dataitems.DistortionBase import DistortionBase
from micado.dataitems.DistortionElt import DistortionElt
from micado.dataitems.DistortionRefBase import DistortionRefBase
from micado.dataitems.DistortionWam import DistortionWam
from micado.dataitems.DpsicsStaroffsetOpt import DpsicsStaroffsetOpt  # noqa: F401
from micado.dataitems.DpsicsStaroffsetSci import DpsicsStaroffsetSci
from micado.dataitems.IllumCorr import IllumCorr
from micado.dataitems.MasterDark import MasterDark
from micado.dataitems.MasterFlat import MasterFlat
from micado.dataitems.Nonlinearity import Nonlinearity
from micado.dataitems.Persistence import Persistence
from micado.dataitems.PhotomBase import PhotomBase
from micado.dataitems.PhotomRef import PhotomRef
from micado.dataitems.PhotomStdfield import PhotomStdfield
from micado.dataitems.DpsicsStaroffsetBase import DpsicsStaroffsetBase
from micado.dataitems.Stack import Stack
from micado.dataitems.StarcenterBase import StarcenterBase
from micado.dataitems.StarcenterWaffle import StarcenterWaffle
from micado.dataitems.Staroffset import Staroffset
from micado.dataitems.DepersistedBase import DepersistedBase
from micado.dataitems.Wam import Wam
from micado.dataitems.WamRef import WamRef
from micado.main.Cat import Cat
from micado.main.DataItem import DataItem
from micado.main.DataItemExtra import (
    KEY_SCI,
    KEY_DQ,
    KEY_ERR,
    KEY_CNT,
    KEY_PERS,
    KEY_BG,
    KEY_CAT,
    KEY_TAB,
    KEY_DATAN,
    KEY_CHI2,
    KEY_DOF,
)
from micado.main.Ext import Ext
from micado.main.Img import Img
from micado.main.Opt import Opt
from micado.main.Raw import Raw
from micado.main.Recipe import Recipe
from micado.main.RecipeStepExtra import RecipeStepBase, name_eso_from_property
from micado.main.Sky import Sky
from micado.main.SkyHCIOnline import SkyHCIOnline
from micado.main.Tab import Tab
from micado.main.Template import Template
from micado.recipes.McdHciOptSciBase import McdHciOptSciBase
from micado.recipes.McdImgCatalogableBase import McdImgCatalogableBase
from micado.recipes.McdImgDetrendHCI import McdImgDetrendHCI
from micado.templates.MicadoCoroAcqStar import MicadoCoroAcqStar
from micado.toolbox.ditaa_from_wise import get_dataitems_derived
from micado.toolbox.rst_dataitemstable_from_python import (
    CLASSES_EXTERNAL,
    CLASSES_RAW,
    # TODO: Find better way to include/exclude DPSICS libraries
    # CLASSES_ICS,
)

from micado.toolbox.create_associationmatrix import (
    CLASSES_STANDARD_IMAGING,
    CLASSES_ASTROMETRIC_IMAGING,
    CLASSES_HIGH_CONTRAST_IMAGING,
    CLASSES_QACITS,
)

from micado.toolbox.rst_from_python import (
    PATH_DOCS_GENERATED,
    CLASSES_DATAITEMS,
    key_input_products,
    key_output_products,
    key_qc_parameters,
    key_accuracy,
    key_user_parameters,
    key_recipeid,
    key_recipe_inputto,
    key_recipe_outputof,
    key_name,
    key_description,
    key_procatg,
    key_processparas,
    key_docatg,
    key_template,
    key_dpr,
    key_otherkeywords,
    key_oca,
    key_structure,
    key_processing,
    dictionary_from_dataitem,
    key_purpose,
    key_algorithm,
    key_errorconditions,
    key_recipetype,
    key_objective,
    key_procedure,
    key_templatename,
    key_recipename,
    keywords_from_qc,
    key_responsible,
    key_phase,
    key_micadosetup,
    key_calibunitsetup,
    key_templateparams,
    key_observingblock_name,
    key_observation_duration,
    key_observation_frequency,
    key_prerequisites,
    key_processed_data,
    key_raw_data,
    key_parent_reqid,
    key_reqid,
    key_remarks,
    key_recipename_cp,
    inputdata_from_dataitem,
    parameters_from_template,
)

from micado.toolbox.augment_backward_chaining import augment_backward_chaining

# Put some keywords in variables to prevent reusing them.
# TODO: Refactor this all, e.g. put all this in MObject or so.
from micado.toolbox.tsf_from_python import get_template_classes

# Import all classes to fill classcache.
# noinspection PyUnresolvedReferences
import micado.main.aweimports  # noqa NOSONAR
from micado.util.esokeywords import esokeywords

CLASSES_CALIBRATION_PLAN = [
    # Astrom,
    IllumCorr,
    MasterDark,
    MasterFlat,
    # Persistence,
    Nonlinearity,
    # TODO: List PhotomBase here and fetch -StdField and -Science?
    PhotomStdfield,
    Staroffset,
    # StarcenterWaffle,
    DistortionElt,
    DistortionWam,
]

CLASSES_STANDARD_IMAGING = CLASSES_STANDARD_IMAGING

# These should have their own pipelines.
CLASSES_WITHOUT_PIPELINE = [
    Persistence,
    Nonlinearity,
    IllumCorr,
    DistortionElt,
]

CLASSES_INTERMEDIATE_DRL_FUNCTION = [
    Img,
    DetrendFinal,
    PhotomBase,
    StarcenterBase,
    DepersistedDark,
    DepersistedNonlin,
    DistortionRefBase,
    DistortionBase,
]

RECIPES_ALL = [
    v.aclass
    for v in classcache.values()
    if issubclass(v.aclass, Recipe) and v.aclass not in [Recipe, McdImgCatalogableBase]
    # TODO: Find better way to include/exclude DPSICS libraries
    and not issubclass(v.aclass, DpsicsStaroffsetBase.recipe.prop_type)
]


DATAITEMS_ALL = [
    v.aclass for v in classcache.values() if issubclass(v.aclass, DataItem)
]


def pipelines_from_dataitem(di):
    """Get pipelines from dataitem."""
    pipelines = []
    if any(issubclass(c, di) for c in CLASSES_STANDARD_IMAGING):
        pipelines.append(("Standard Imaging", "fig:associationmatrixsi"))
    if any(issubclass(c, di) for c in CLASSES_ASTROMETRIC_IMAGING):
        pipelines.append(("Astrometric Imaging", "fig:associationmatrixai"))
    if any(issubclass(c, di) for c in CLASSES_HIGH_CONTRAST_IMAGING):
        pipelines.append(("High Contrast Imaging", "fig:associationmatrixhci"))
    if any(issubclass(c, di) for c in CLASSES_QACITS):
        pipelines.append(("Online HCI", "fig:associationmatrixqacits"))
    return pipelines


@dataclass
class ParameterInfo:
    """Info about parameter."""

    # TODO: Refactor this into attributes of persistent properties or so.
    name: str = "None"
    type: str = "None"
    default: str = "None"
    description: str = "None"
    name_eso: str = "None"
    unit: str = "None"

    @classmethod
    def from_property(cls, prop):
        """Create ParameterInfo from persistent property."""
        desc, _, unit = (a.strip(" ]") for a in prop.__doc__.split("["))
        nameeso = name_eso_from_property(prop)
        # TODO: or get this from eso_keywords?
        return ParameterInfo(
            name=prop.attribute.upper(),
            type=type(prop.default).__name__,
            default=str(prop.default),
            description=desc,
            name_eso=nameeso,
            unit=unit,
        )


def get_dataitems_to_inherit_kws_from(di):
    """E.g. DepersistedDark inherits kws from Dark.

    So the kws from DepersistedDark that are used in OCA rules, should be
    present in Dark as well, and are used in OCA rules indirectyl.
    """
    yield di
    r = di.get_progenitor_class()
    if r:
        yield from get_dataitems_to_inherit_kws_from(r)


def get_oca_keywords_full():
    """Get full set of oca keywords per dataitem."""
    # Add the backward chaining.
    augment_backward_chaining()
    # Get all dataitems of interest.
    dataitems = [
        v.aclass
        for v in classcache.values()
        if issubclass(v.aclass, (Raw, Img, Ext, Tab, Cat))
        and not issubclass(v.aclass, RecipeStepBase)
    ]
    properties_used_in_lineage = [
        left_or_right
        for di in dataitems
        for condition in di.lineage.values()
        for left, _, right in condition
        for left_or_right in (left, right)
    ]
    # Get all (class, fitsname) for properties used in lineage.
    cls_fits_set = {
        (p.cls, name_eso_from_property(p)) for p in properties_used_in_lineage
    }
    cls_fits = sorted(cls_fits_set, key=lambda c_a: (c_a[0].__name__, c_a[1]))
    # Get all attributes that are used for each class in OCA rules.
    fits_oca_from_class = {
        c: [a for _, a in c_attrs]
        for c, c_attrs in itertools.groupby(cls_fits, key=lambda a_c: a_c[0])
    }
    # Get the progenitor classes that DataItems inherit FITS kws from.
    diss_to_inherit = [
        (di_sup, di)
        for di in dataitems
        for di_sup in get_dataitems_to_inherit_kws_from(di)
        if not di_sup == di
    ]
    diss_to_inherit.sort(
        key=lambda progenitor_product: [c.__name__ for c in progenitor_product]
    )
    # Invert them, list the DataItems that are derived from a DataItem.
    derived_from_di = {
        progenitor: [product for _, product in products]
        for progenitor, products in itertools.groupby(
            diss_to_inherit,
            key=lambda prog_prod: prog_prod[0],
        )
    }
    # Sanity check.
    assert not set(derived_from_di) - set(dataitems)
    # Get the full set of attributes of a class that are used in OCA rules,
    # also the attributes that are only used in OCA rules of derived classes.
    fits_oca_from_class_full = {
        di: sorted(
            set(
                k.replace(" ", ".")
                # TODO: Use FILTi.NAME directly in the OCA rules.
                .replace("FILT.NAME", "FILTi.NAME")
                # TODO: Fix READMODE in oca rules
                .replace("INS.READMODE", "DET.READ.MODE")
                for k in fits_oca_from_class.get(di, [])
                + [
                    attr
                    for prog in derived_from_di.get(di, [])
                    for attr in fits_oca_from_class.get(prog, [])
                ]
            )
        )
        for di in dataitems
    }
    return fits_oca_from_class_full


FITS_OCA_FROM_CLASS_FULL = get_oca_keywords_full()


def latex_from_string(s):
    """Create LaTeX string."""
    return (
        s.strip()
        .replace("_", "\\_")
        .replace("&", "\\&")
        .replace("%", "\\%")
        .replace("#", "\\#")
        # trailing space is necessary
        .replace("\n", "\\newline ")
        # TODO: Fix this \\cite hack.
        .replace("\\\\cite", "\\cite")
    )


def latex_from_string_list(s: List[str], add_linebreaks=False) -> str:
    """Create LaTeX string."""
    snew = [latex_from_string(line) for line in s]
    sjoin = " \\newline\n    " if add_linebreaks else " \\newline "
    return sjoin.join(snew)


def get_simple(dt, kk):
    """Get simple key, value pair for table."""
    return kk, latex_from_string_list(dt[kk])


def latex_label_section_keyword(name):
    """Get label from latex keyword."""
    name = name.replace("HIERARCH ESO ", "")
    label_section = name.replace("i", "").lower().replace(" ", ".").replace("_", "")
    if name in ["NAXIS1", "NAXIS2"]:
        label_section = "NAXISn".lower()
    elif ".sur." in label_section:
        # TODO: Fix this hack for surfaces.
        # e.g. ins.sta.sur.name or rel.sur.sur.outer
        # -> sur.name and sur.outer
        label_section = ".".join(label_section.split(".")[2:])
    elif ".det." in label_section:
        # TODO: Fix this hack for detectors.
        # e.g. det.det.angle -> det.angle
        label_section = ".".join(label_section.split(".")[1:])
    else:
        # Replace e.g. TPLi START with just TPL.START
        label_section = name.replace("i", "").lower().replace(" ", ".").replace("_", "")
    return label_section


def latex_keyword(name, link=True, qc=False):
    """Create LaTeX version of keyword."""
    if name in ["N/A", "None"]:
        return name
    name = name.replace("HIERARCH ESO ", "")
    name = name.replace(" ", ".")
    label_section = latex_label_section_keyword(name)
    smacro = "QC" if qc else "FITS"
    if name.lower() in ["none", "todo"]:
        link = False
    return (
        f"\\hyperref[parameter:{label_section}]{{\\{smacro}{{{name}}}}}"
        if link
        else f"\\{smacro}{{{name}}}"
    )


def latex_dataitem(name, link=True):
    """Create LaTeX version of name of DataItem."""
    label_section = name.replace("_", "").replace("\\", "")
    return (
        f"\\hyperref[dataitem:{label_section}]{{\\PROD{{{name}}}}}"
        if link
        else f"\\PROD{{{name}}}"
    )


def latex_template(name, link=True):
    """Create LaTeX version of name of a template."""
    label_section = (
        name.replace("_", "").replace("\\", "").replace("[", "").replace("]", "")
    )
    if name in ["N/A"]:
        return "N/A"
    return (
        f"\\hyperref[template:{label_section}]{{\\TPL{{{name}}}}}"
        if link
        else f"\\TPL{{{name}}}"
    )


def latex_recipe(name, link=True, include_normal_reference=False):
    """Get LaTeX name for recipe."""
    if name == "None":
        return "None"
    label_section = (
        name.replace("_", "")
        .replace("\\", "")
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
    )
    normal_reference = (
        f""" (\\ref{{recipe:{label_section}}})""" if include_normal_reference else ""
    )
    return (
        f"\\hyperref[recipe:{label_section}]{{\\REC{{{name}}}}}{normal_reference}"
        if link
        else f"\\REC{{{name}}}"
    )


def latex_recipestep(name, link=True, include_normal_reference=False):
    """Get LaTeX name for recipe."""
    label_section = (
        name.replace("_", "")
        .replace("\\", "")
        .replace(" ", "")
        .replace(":", "")
        .replace(".", "")
    )
    # TODO: replace 'dataitem' with 'recipestep' or so.
    normal_reference = (
        f""" (\\ref{{dataitem:{label_section}}})""" if include_normal_reference else ""
    )
    return (
        f"\\hyperref[dataitem:{label_section}]{{\\DRL{{{name}}}}}{normal_reference}"
        if link
        else f"\\DRL{{{name}}}"
    )


def get_durations_from_dataitem(dataitem):
    """Get the duration of a dataitem.

    It used to be that the duration is the duration of the template. Now it is
    the duration of the OB. Therefore the duration of MasterFlat and
    Nonlinearity differ, even though they use just Flat as law.
    """
    raw1 = dataitem.get_raw_class()
    # TODO: fix hack
    if not isinstance(raw1, tuple):
        raw1 = (None, raw1)
    _, raw = raw1
    if dataitem.text_pro.duration:
        duration = dataitem.text_pro.duration
    elif raw.text_raw.duration:
        duration = raw.text_raw.duration
    else:
        duration = "TBD"

    durationper24hrs = (
        str(raw.text_raw.durationper24hrs)
        if raw.text_raw.durationper24hrs != "None"
        else duration
    )

    return duration, durationper24hrs


def get_tex_calibration_summary_row(dataitem, withaverageduration=False):
    """Get row for Calibration Plan Summary table."""
    assert dataitem in CLASSES_CALIBRATION_PLAN
    recipe = dataitem.recipe.prop_type
    raw = dataitem.get_raw_class()
    template = raw.template.prop_type
    duration, durationper24hrs = get_durations_from_dataitem(dataitem)
    # duration = duration.split()[0]
    text_phase = raw.text_raw.phase or "?"
    if withaverageduration:
        text_duration = f"{duration}".split()[0]
        text_durationper24hrs = f"{durationper24hrs} ({text_phase[0]})"
    else:
        text_duration = f"{duration} ({text_phase[0]})"
        text_durationper24hrs = f"{duration}"
    cells = (
        [
            latex_from_string(dataitem.text_pro.name),
            latex_template(latex_from_string(template.name_eso()), link=False),
            text_duration,
        ]
        + ([text_durationper24hrs] if withaverageduration else [])
        + [
            latex_dataitem(raw.name_eso().strip("_"), link=False),
            latex_recipe(latex_from_string(recipe.name_eso()), link=False),
            # TODO: Add latex_from_string here too.
            latex_dataitem(dataitem.name_eso()),
        ]
    )
    return "    " + " & \n    ".join(cells) + " \\tabularnewline\n    \\hline\n"


def get_parameters_full(dataitem):
    """Get full set of parameters for dataitem."""
    # TODO: refactor
    assert hasattr(
        dataitem, "recipe"
    ), f"{dataitem.__name__} does not have a Recipe associated!"
    recipe = dataitem.recipe.prop_type
    return recipe.get_parameters_c()


def get_structure(dataitem: Type[DataItem]) -> str:
    """Structure text."""
    # TODO: Support optional layers.
    s_layers = ",".join(f"\\FITS{{{ln}}}" for ln in dataitem.layer_names()[0])
    s_each = "each " if len(dataitem.layer_names()[0]) > 1 else ""
    lmin, lmax = dataitem.layers_number()
    is_virtual = issubclass(dataitem, (DpsicsStaroffsetBase, SkyHCIOnline))
    if lmin is None and lmax is None:
        s_num = "for each exposure"
        raise ValueError(s_num)
    elif lmin is None and lmax == 1:
        s_num = s_each + "at most once"
        raise ValueError(s_num)
    elif lmin is None:
        s_num = s_each + f"at most {lmax} times"
        raise ValueError(s_num)
    elif lmin == 1 and lmax is None:
        assert issubclass(dataitem, CubeHCI)
        s_num = s_each + "once for each exposure"
    elif lmax is None:
        s_num = f"each at least {lmin} times"
        raise ValueError(s_num)
    elif lmin == lmax == 1:
        # Explicitely ensure the background is now per detector.
        assert not issubclass(dataitem, Background), dataitem
        if issubclass(dataitem, Stack):
            s_num = s_each + "once, spanning all detectors"
        elif issubclass(dataitem, DistortionBase):
            # TODO: Determine whether this is actually true.
            s_num = s_each + "once, spanning all detectors"
        elif issubclass(dataitem, Tab):
            assert issubclass(
                dataitem, (DpsicsStaroffsetBase, StarcenterBase, Staroffset)
            ), dataitem
            s_num = "once, for the central detector"
        elif issubclass(dataitem, Cat):
            s_num = "once"
        else:
            s_num = s_each + "once, for the central detector"
    elif lmin == lmax:
        s_num = s_each + f"once for each detector, so {lmin} times"
    else:
        assert lmin == 1 and lmax == 9
        s_num = s_each + f"once for each detector, so {lmin} or {lmax} times"
    s_drl_l = f'\\ac{{DRL}} structure in \\ref{{drsstructure:{dataitem.name_eso(includebase=False).replace("_", "")}}}'
    # Headers link is by default removed.
    # s_headers_link = f'headers in \\ref{{fitskeywords:{
    #     dataitem.name_eso(includebase=True).replace("_", "")}}}'
    # s_text = f"{s_layers}; {s_num}; {s_drl_l}; {s_headers_link}"
    s_text = f"{s_layers}; {s_num}; {s_drl_l}"
    return s_text if not is_virtual else f"({s_text})"


def get_keywords_processing(dataitem: Type[DataItem]) -> List[str]:
    """Get FITS keywords for processing."""
    fits_processing = []
    # TODO: Fix hack that SkyHCIOnline does not get astrometry headers.
    if not dataitem == SkyHCIOnline:
        for baseclass in dataitem.__bases__:
            fits_processing += get_keywords_processing(baseclass)
    if hasattr(dataitem, "text_raw"):
        fits_processing += dataitem.text_raw.keywordsprocessing
    if hasattr(dataitem, "text_pro"):
        fits_processing += dataitem.text_pro.keywordsprocessing

    # Make list unique while preserving order.
    fits_processing = list(dict.fromkeys(fits_processing))
    return fits_processing


def get_keywords_structured(dataitem: Type[DataItem]) -> Tuple[List[str]]:
    """Get FITS keywords in a structured way."""
    fits_all = dataitem.fits_keywords_c()
    fits_filename = [k for k in fits_all if "FILENAME" in k]
    fits_hierarch = [
        k.replace("HIERARCH ESO ", "") for k in fits_all if k not in fits_filename
    ]
    fits_hierarch_qc = [
        k.replace(" ", ".") for k in fits_hierarch if k.startswith("QC ")
    ]
    fits_hierarch_rec = [
        k.replace(" ", ".") for k in fits_hierarch if k.startswith("REC ")
    ]
    fits_hierarch_other = [
        k.replace(" ", ".")
        for k in fits_hierarch
        if k.replace(" ", ".") not in fits_hierarch_rec + fits_hierarch_qc
    ]
    fits_other = [
        k.replace(" ", ".")
        for k in fits_all
        if k.replace("HIERARCH ESO ", "") not in fits_hierarch
        and k not in fits_filename
    ]
    fits_oca = FITS_OCA_FROM_CLASS_FULL[dataitem]
    fits_processing = get_keywords_processing(dataitem)
    return (
        fits_all,
        fits_hierarch_qc,
        fits_hierarch_rec,
        fits_hierarch_other,
        fits_oca,
        fits_other,
        fits_processing,
    )


def get_essential_raw_keywords():
    """All essential keywords for data processing."""
    kws_all = [
        "DPR.CATG",
        "DPR.TYPE",
        "DPR.TECH",
        "TPL.EXPNO",
        "OBS.TPLNO",
        "OCS.PXSCALE",
        # TODO: OPTI6 is pupil wheel and should be automatic.
        "INS.OPTI6.NAME",
    ]
    for dataitem in CLASSES_RAW:
        (
            fits_all,
            fits_hierarch_qc,
            fits_hierarch_rec,
            fits_hierarch_other,
            fits_oca,
            fits_other,
            fits_processing,
        ) = get_keywords_structured(dataitem)
        assert not fits_hierarch_qc, f"Should not be QC params: {fits_hierarch_qc}"
        assert not fits_hierarch_rec, f"Should not be REC params: {fits_hierarch_rec}"
        kws = fits_oca + fits_processing
        kws_all += kws

    kws_all_sorted = sorted(set(kws_all))
    # TODO: Make a proper test out of this.
    kws_all_expected = [
        "CDELTn",
        "CRPIXn",
        "CRVALn",
        "CTYPEn",
        "CUNITn",
        "DET.DIT",
        "DET.READ.MODE",
        "DPR.CATG",
        "DPR.TECH",
        "DPR.TYPE",
        "EXPTIME",
        "INS.FILTi.NAME",
        "INS.OPTI6.NAME",
        "INSTRUME",
        "MJD-OBS",
        "OBS.TPLNO",
        "OCS.PXSCALE",
        "TPL.EXPNO",
        "TPL.START",
    ]

    assert kws_all_expected == kws_all_sorted
    return kws_all_sorted


def get_tex_essential_raw_keywords():
    """Get tech list of all essential RAW FITS keywords."""
    kws = get_essential_raw_keywords()
    s_kws = "".join(
        [f"    \\item {latex_from_string(latex_keyword(p))}\n" for p in kws]
    )
    s_tex = f"""\\begin{{itemize}}
    \\setlength{{\\itemsep}}{{0pt}}
    {s_kws}
\\end{{itemize}}"""
    return s_tex


def get_progenitors_for_recipe(recipe):
    """Get all possible progenitors of recipe.

    This cannot actually be done on the recipe level, but only on the processed
    dataitem level. That is because a recipe does not actually specify its
    input, only the previous step. Which describes the previous, etc.

    In particular, the first step, that has the progenitor of the processed
    data product as input, will probably have a very generic input data item.

    E.g. IllumCorr is created from DetrendIllum through McdImgIllum. However,
    the fist step of McdImgIllum is SubtractBackground, which has only
    DetrendBase as input. So McdImgIllum does not know it needs a DetrendIllum.
    """
    dataitemswrecipes = list(
        set(
            di
            for di in DATAITEMS_ALL
            if hasattr(di, "recipe") and not di.__name__.endswith("Base")
        )
    )
    return [
        di.get_progenitor_class()
        for di in dataitemswrecipes
        if issubclass(di.recipe.prop_type, recipe)
    ]


def get_possible_input_for_recipe(recipe):
    """Get the full possible input for this recipe.

    See get_progenitors_for_recipe() for rationale.
    """
    return get_progenitors_for_recipe(recipe) + recipe.get_inputdata()[1:]


def get_recipes_that_use_dataitem(dataitem):
    """Get the recipes that use this dataitem as input.

    This is complicated due to many hacks.
    """
    recipes_not_base = [
        recipe for recipe in RECIPES_ALL if not recipe.name_eso().endswith("base")
    ]
    recipes = [
        recipe for recipe in recipes_not_base if dataitem in recipe.get_inputdata()
    ]

    # Add e.g. McdImgDetrendFinal to Detrend, DetrendAi,
    #          McdImgPhotom to DetrendFinal,
    #          McdImgDistortion to DetrendFinalAstromfield and DetrendFinalWam
    recipes += [
        recipe
        for recipe in recipes_not_base
        if dataitem in get_progenitors_for_recipe(recipe) and recipe not in recipes
    ]

    # Add McdImgBackground to DetrendCoroBackground, Detrend, DetrendAi,
    #     McdImgAstrom, McdImgCalibrate to DetrendFinal
    recipes += [
        recipe
        for recipe in recipes_not_base
        if any(
            dataitem in di.get_derived_classes()
            for di in get_progenitors_for_recipe(recipe)
        )
        and recipe not in recipes
    ]

    # Add e.g. McdImgCalibHCI to StarCenterWaffle and StarCenterPupil
    #          McdImgStarPhotom, McdImgCalibrate to PhotomScience, PhotomStdfld
    recipes += [
        recipe
        for recipe in recipes_not_base
        if any(
            dataitem in di.get_derived_classes()
            for di in get_possible_input_for_recipe(recipe)
        )
        and recipe not in recipes
    ]

    # Add various recipes to Base classes. Need to use issubclass here and
    # not "in di.get_derived_classes()
    # because otherwise DetrendBkgfield can be used in McdImgDetrendFinal
    # TODO: Fix this Background hack
    recipes += [
        recipe
        for recipe in recipes_not_base
        if any(issubclass(di, dataitem) for di in get_possible_input_for_recipe(recipe))
        and recipe not in recipes
    ]

    return recipes


def get_tex_contents_dataitem_pro(dataitem):
    """Contents of tex file for processed dataitem."""
    (
        fits_all,
        fits_hierarch_qc,
        fits_hierarch_rec,
        fits_hierarch_other,
        fits_oca,
        fits_other,
        fits_processing,
    ) = get_keywords_structured(dataitem)
    fits_hierarch_qc = fits_hierarch_qc or ["None"]
    recipes = get_recipes_that_use_dataitem(dataitem)
    dic1 = {
        key_name: dataitem.name_eso(),
        key_recipe_outputof: dataitem.recipe.prop_type.name_eso(),
        key_description: dataitem.text_pro.objective,
        key_procatg: dataitem.name_oca(),
        key_qc_parameters: fits_hierarch_qc,
        key_processparas: fits_hierarch_rec,
        key_otherkeywords: fits_hierarch_other + fits_other,
        key_oca: fits_oca,
        key_processing: fits_processing,
        key_recipe_inputto: [r.name_eso() for r in recipes] if recipes else ["None"],
    }
    dic3 = {
        key_name: latex_dataitem(latex_from_string(dic1[key_name])),
        key_description: latex_from_string(dic1[key_description]),
        key_oca: latex_from_string_list([latex_keyword(p) for p in dic1[key_oca]]),
        key_procatg: f"\\FITS{{{latex_from_string(dic1[key_procatg])}}}",
        key_recipe_outputof: latex_recipe(latex_from_string(dic1[key_recipe_outputof])),
        key_recipe_inputto: latex_from_string_list(
            [latex_recipe(r) for r in dic1[key_recipe_inputto]]
        ),
        key_qc_parameters: latex_from_string_list(
            [latex_keyword(p, qc=True) for p in fits_hierarch_qc]
        ),
        key_processing: latex_from_string_list(
            [latex_keyword(p) for p in dic1[key_processing]]
        )
        if dic1[key_processing]
        else "None",
        key_structure: get_structure(dataitem),
        # key_processparas: latex_from_string_list(
        #     [f"\\PAR{{{p}}}" for p in dic1[key_processparas]]
        # ),
        # key_otherkeywords: latex_from_string_list(
        #     [f"\\PAR{{{p}}}" for p in dic1[key_otherkeywords]]
        # ),
    }
    lines_table = [f"{name}: & " + ls + "\\\\" for name, ls in dic3.items()]
    stable = "\n".join(lines_table)
    label_section = dataitem.name_eso().replace("_", "")
    name_section = latex_from_string(dataitem.name_eso())
    return f"""\\paragraph{{{name_section}}}\\label{{dataitem:{label_section}}}
\\begin{{recipedef}}
{stable}
\\end{{recipedef}}
"""


def rebreak_lines(name, ls):
    """Rebreak lines that got joined in latex_from_string_list..."""
    ls2 = ls.split(" \\newline ")
    line0 = f"{name}: & " + ls2[0] + "\\\\"
    lines = [f" & {line} \\\\" for line in ls2[1:]]
    return [line0] + lines


def get_tex_contents_dataitem_raw(dataitem, recipes: List[Type[Recipe]]):
    """Contents of tex file for raw of dataitem."""
    (
        fits_all,
        fits_hierarch_qc,
        fits_hierarch_rec,
        fits_hierarch_other,
        fits_oca,
        fits_other,
        fits_processing,
    ) = get_keywords_structured(dataitem)
    dpr = [dataitem.dpr_catg_c(), dataitem.dpr_type_c(), dataitem.dpr_type_c()]
    # TODO: Split up the recipes in different types? Sort differently?
    recipes.sort(key=lambda r: r.name_eso())
    recipe_names = [recipe.name_eso() for recipe in recipes]
    dic1 = {
        key_name: dataitem.name_eso(),
        # This recipe is unstable, depending on the order classes are imported.
        # TODO: Make this always the main recipe, if there is one.
        key_recipe_inputto: recipe_names,
        key_description: dataitem.description_oneline(),
        key_dpr: dpr,
        key_docatg: dataitem.name_oca(),
        key_template: dataitem.template.prop_type.name_eso(),
        key_otherkeywords: fits_hierarch_other + fits_other,
        key_oca: fits_oca,
        key_processing: fits_processing,
    }
    dic3 = {
        key_name: latex_from_string(latex_dataitem(dic1[key_name])),
        key_description: latex_from_string(dic1[key_description]),
        key_dpr: latex_from_string_list([f"\\FITS{{{k}}}" for k in dic1[key_dpr]]),
        key_oca: latex_from_string_list([latex_keyword(p) for p in dic1[key_oca]]),
        key_docatg: f"\\FITS{{{latex_from_string(dic1[key_docatg])}}}",
        key_template: latex_template(latex_from_string(dic1[key_template])),
        # TODO: Move other keywords to inheritance appendix?
        # key_otherkeywords: latex_from_string_list(
        #     [f"\\PAR{{{p}}}" for p in dic1[key_otherkeywords]]
        # ),
        key_recipe_inputto: latex_from_string_list(
            [latex_recipe(recipe_name) for recipe_name in dic1[key_recipe_inputto]]
        ),
        key_processing: latex_from_string_list(
            [latex_keyword(p) for p in dic1[key_processing]]
        )
        if dic1[key_processing]
        else "None",
        key_structure: get_structure(dataitem),
    }
    lines_table = [
        line for name, ls in dic3.items() for line in rebreak_lines(name, ls)
    ]
    stable = "\n".join(lines_table)
    label_section = dataitem.name_eso().replace("_", "")
    name_section = latex_from_string(dataitem.name_eso())
    return f"""\\paragraph{{{name_section}}}\\label{{dataitem:{label_section}}}
\\begin{{recipedef}}
{stable}
\\end{{recipedef}}
"""


def get_tex_contents_dataitem_ext(dataitem):
    """Contents of tex file for external of dataitem."""
    (
        fits_all,
        fits_hierarch_qc,
        fits_hierarch_rec,
        fits_hierarch_other,
        fits_oca,
        fits_other,
        fits_processing,
    ) = get_keywords_structured(dataitem)
    recipes = get_recipes_that_use_dataitem(dataitem)

    dic1 = {
        key_name: dataitem.name_eso(),
        key_recipe_inputto: [r.name_eso() for r in recipes],
        key_description: dataitem.description_oneline(),
        key_docatg: dataitem.name_oca(),
        key_otherkeywords: fits_hierarch_other + fits_other,
        key_oca: fits_oca,
    }
    dic3 = {
        key_name: latex_from_string(latex_dataitem(dic1[key_name])),
        key_description: latex_from_string(dic1[key_description]),
        key_oca: latex_from_string_list([latex_keyword(p) for p in dic1[key_oca]]),
        key_docatg: f"\\FITS{{{latex_from_string(dic1[key_docatg])}}}",
        key_recipe_inputto: latex_from_string_list(
            [latex_recipe(r) for r in dic1[key_recipe_inputto]]
        ),
        # key_otherkeywords: latex_from_string_list(
        #     [f"\\PAR{{{p}}}" for p in dic1[key_otherkeywords]]
        # ),
        key_processing: "TODO",
        key_structure: get_structure(dataitem),
    }
    lines_table = [
        line for name, ls in dic3.items() for line in rebreak_lines(name, ls)
    ]
    stable = "\n".join(lines_table)
    label_section = dataitem.name_eso().replace("_", "")
    name_section = latex_from_string(dataitem.name_eso())
    return f"""\\paragraph{{{name_section}}}\\label{{dataitem:{label_section}}}
\\begin{{recipedef}}
{stable}
\\end{{recipedef}}
"""


def write_datatitem_external_tex(dataitem: DataItem):
    """Write an external dataitem to a tex file."""
    fname = dataitem.name_file_doc()
    stex_pro = get_tex_contents_dataitem_ext(dataitem)
    fn_rst = PATH_DOCS_GENERATED / f"{fname}.tex"
    with open(fn_rst, "w") as f:
        f.write(stex_pro)


def get_tex_input_products(dataitem, link=True, allowed=None):
    """Get the tex for the input products of dataitem."""
    inputdata, inputdata_multi = inputdata_from_dataitem(dataitem, allowed=allowed)
    s_input_products = latex_from_string_list(
        [
            latex_dataitem(inpu.name_eso().strip("_"), link=link)
            + (
                (
                    " ("
                    + " or ".join(
                        [
                            latex_dataitem(p.name_eso().strip("_"), link=link)
                            for p in inputdatas
                        ]
                    )
                    + ")"
                )
                if inputdatas not in [[], [inpu]]
                else ""
            )
            for inpu, inputdatas in zip(inputdata, inputdata_multi)
        ],
        add_linebreaks=True,
    )
    return s_input_products


def get_tex_output_product(dataitem, link=True, allowed=None):
    """Get the tex of the output product, that is, the dataitem itself."""
    # Strict must be True for DetrendBase because DetrendPupil is a 'derived'
    # class of DetrendBkgfield. TODO fix that Background hack.
    dataitems_derived = get_dataitems_derived(dataitem, allowed=allowed, strict=True)
    s_output_product = latex_from_string(
        latex_dataitem(dataitem.name_eso().strip("_"), link=link)
        + (
            (
                " ("
                + " or ".join(
                    [
                        latex_dataitem(p.name_eso().strip("_"), link=link)
                        for p in dataitems_derived
                    ]
                )
                + ")"
            )
            if dataitems_derived not in [[], [dataitem]]
            else ""
        )
    )
    return s_output_product


def get_tex_contents_recipe(dataitem, dataitems):
    """Create contents of tex file for recipe.

    See rule 1387 of ESO-037611 v4
    Dataflow_for_ESO_Observatories_Deliverables_Standard
    """
    recipe = dataitem.recipe.prop_type
    params = [pn for pn in get_parameters_full(dataitem)]

    dataitems_derived = get_dataitems_derived(dataitem)
    # TODO: Fix this DetrendBase mess
    if not dataitem == DetrendBase:
        assert not set(dataitems_derived) - set(dataitems)
    can_be_base = set(dataitems) - set(dataitems_derived)
    assert all("Base" in di.__name__ for di in can_be_base)

    # TODO: Refactor this allowed stuff with create_associationmatrix
    if dataitem in (DetrendFinalBase, DetrendBase):
        allowed_input = tuple(di.get_progenitor_class() for di in dataitems_derived)
    else:
        allowed_input = None

    # TODO: Refactor this allowed stuff with create_associationmatrix
    if dataitem == DetrendBase:
        output_hci = get_dataitems_derived(DetrendHCIBase)
        allowed_output = [di for di in dataitems_derived if di not in output_hci]
    else:
        allowed_output = None

    # TODO: dataitems is unused, why?
    dataitems.sort(key=lambda di: di.name_eso())
    s_input_products = get_tex_input_products(dataitem, allowed=allowed_input)

    recipesteps = recipe.get_all_steps()
    names_steps = [latex_recipestep(rs.name_eso()) for rs in recipesteps]
    # TODO: decide to not include ["See individual DRL functions:"] +
    algorithm = names_steps
    st_parameters = (
        latex_from_string_list(
            [
                f"\\PAR{{{name_eso_from_property(prop)}}} ({prop.default})"
                for prop in params
            ]
        )
        if params
        else "None"
    )

    # Get QC parameters
    qc = dataitem.get_qc_class()
    lqc = keywords_from_qc(qc) if qc is not None else ["None"]

    raw = dataitem.get_raw_class()
    template = raw.template.prop_type
    stype = template.type_eso()
    # Fix unknown types.
    if stype == "Unknown":
        if dataitem in [Persistence, PhotomBase, DistortionBase]:
            stype = "Calibration"
        elif dataitem in [DetrendBase, DetrendFinalBase, DepersistedBase, CalibHCI]:
            stype = "Science/Calibration"
        elif dataitem in [CubeHCI]:
            stype = "Science"
        else:
            # Sanity check to ensure that Unknown types are dealt with
            # explicitly.
            raise ValueError(f"Unsupported: {stype} {dataitem.name_eso()}")
    elif stype == "Technical":
        assert dataitem in [IllumCorr, Nonlinearity]
        stype = "Calibration"

    pipelines = pipelines_from_dataitem(dataitem)
    assert (
        pipelines or dataitem in CLASSES_WITHOUT_PIPELINE
    ), f"No pipelines found for {dataitem}"
    if dataitem in CLASSES_WITHOUT_PIPELINE:
        spipeliness = "None"
    else:
        spipeliness = ", ".join(
            f"{name_pipeline} (Figures~\\ref{{{label_fig}}}~and~\\ref{{{label_fig}big}})"
            for name_pipeline, label_fig in pipelines
        )
    fname = dataitem.name_file_doc()
    sname = latex_from_string(dataitem.recipe.prop_type.name_eso())
    label_name = dataitem.recipe.prop_type.name_eso().replace("_", "")
    label_name_full = f"recipedecomposition:{label_name}"
    sfigure = f"""
\\input{{generated/{fname}_tikz.tex}}
"""
    st_figure_reference = f"See Figure~\\ref{{{label_name_full}}}."

    s_remarks = (
        dataitem.text_pro.remarks.strip() if dataitem.text_pro.remarks else "None"
    )
    s_remarks = s_remarks.replace(
        r" (described in Section~\ref{sec:photometricreferencestars})",
        "",
    )

    dic3 = OrderedDict(
        [
            # The original name is now only used in the Calibration Plan
            # (key_name, latex_from_string(dictable[key_name][0])),
            (
                key_recipename,
                latex_recipe(latex_from_string(recipe.name_eso())),
            ),
            # TODO: Distinguish Calibration and Science recipes
            (key_recipetype, stype),
            # TODO: Decide what to do with Requirements.
            # get_simple(dictable, "ReqID"),
            # get_simple(dictable, "Parent ReqID"),
            # Note that the key changes here because the 611 document
            # prescribes a different key.
            (key_purpose, latex_from_string(dataitem.text_pro.objective)),
            (
                key_input_products,
                s_input_products,
            ),
            # TODO: Add "FITS data structure, data content and identification
            # keywords such as PRO keywords"
            (
                key_output_products,
                get_tex_output_product(dataitem, allowed=allowed_output),
            ),
            (
                key_qc_parameters,
                latex_from_string_list([latex_keyword(p, qc=True) for p in lqc]),
            ),
            (key_user_parameters, st_parameters),
            (key_procedure, st_figure_reference),
            (key_algorithm, latex_from_string_list(algorithm)),
            # TODO: Ensure that these do not contain useful information,
            # since they do not end up anywhere.
            # get_simple(dictable, "Procedure"),
            # get_simple(dictable, key_mathematical_description),
            # TODO: Decide what to do with accuracy?
            # (key_accuracy, latex_from_string(dictable[key_accuracy][0])),
            # ("Verification", text_verification),
            # TODO: Add something about errorconditions
            (key_errorconditions, "None"),
            (
                key_remarks,
                latex_from_string(s_remarks),
            ),
        ]
    )
    lines_table = [f"{name}: & " + ls + "\\\\" for name, ls in dic3.items()]
    stable = "\n".join(lines_table)

    # These lines can be used in case we want to split the figures up again.
    # Both labels are now kept because they might be referenced somewhere.
    # \\subsubsection{{{sname}}}\\label{{recipe:{label_name}}}
    # Recipe decomposition is shown in Section~\\ref{{{label_name_full}}}.
    # \\subsubsection{{{sname}}}\\label{{{label_name_full}}}
    # Recipe is described in Section~\\ref{{recipe:{label_name}}}.

    tex_table = f"""\\subsubsection{{{sname}}}\\label{{recipe:{label_name}}}\\label{{{label_name_full}}}

Used in pipelines: {spipeliness}.

\\begin{{recipedef}}
{stable}
\\end{{recipedef}}
\\clearpage

{sfigure}
\\clearpage
"""
    return tex_table


def get_tex_contents_recipe_calibration_plan(dataitem):
    """Create contents of tex file for recipe."""
    dictable = dictionary_from_dataitem(dataitem)

    raw = dataitem.get_raw_class()
    template = raw.template.prop_type

    use_raw = raw.__name__ != "Science" and dataitem != Persistence
    name_template = template.name_eso() if use_raw else "N/A"

    name_observingblock = dataitem.get_observingblock_name()
    tex_name_observingblock = latex_from_string(name_observingblock)
    if " " not in tex_name_observingblock and tex_name_observingblock != "N/A":
        # Otherwise it is like "
        # this observing Template is part of the science Observing Block"
        tex_name_observingblock = latex_keyword(tex_name_observingblock, link=False)

    qc = dataitem.get_qc_class()
    qc_parameters = keywords_from_qc(qc) if qc else []

    s_input_products = latex_dataitem(raw.name_eso(), link=False)
    s_output_products = get_tex_output_product(dataitem, link=False)

    sname = latex_from_string(dictable[key_name][0])
    label_name_dataitem = dataitem.name_eso().replace("_", "")
    label_name_recipe = dataitem.recipe.prop_type.name_eso().replace("_", "")
    label_table = dataitem.__name__.lower()
    # TODO: These do have input raws.
    # TODO: This dichotomy is not maintainable, because e.g. Detrend
    #       can do many images that need to be described, but not with
    #       the detrend recipe.
    assert len(dictable[key_output_products]) == 1
    dic3 = OrderedDict(
        [
            (key_name, latex_from_string(dictable[key_name][0])),
            get_simple(dictable, key_objective),
            get_simple(dictable, key_responsible),
            get_simple(dictable, key_phase),
            get_simple(dictable, key_micadosetup),
            get_simple(dictable, key_calibunitsetup),
            (
                key_templatename,
                latex_template(latex_from_string(name_template), link=False),
            ),
            (
                key_templateparams,
                latex_from_string_list(
                    [
                        f"\\PAR{{{p}}}" if p != "N/A" else "N/A"
                        for p in dictable[key_templateparams][0].split(", ")
                    ]
                ),
            ),
            # TODO: Add OB to the model?
            # Currently the name is just the template.
            # However the same template can be used in different pre-defined OBs
            (
                key_observingblock_name,
                tex_name_observingblock,
            ),
            get_simple(dictable, key_observation_frequency),
            (
                key_observation_duration,
                str(get_durations_from_dataitem(dataitem)[0]),
            ),
            get_simple(dictable, key_prerequisites),
            # Rename input_products to raw_data.
            (
                key_raw_data,
                s_input_products,
            ),
            (
                key_recipename_cp,
                latex_recipe(latex_from_string(dictable[key_recipeid][0]), link=False),
            ),
            # Rename output_products to processed_data.
            (
                key_processed_data,
                s_output_products,
            ),
            (
                key_qc_parameters,
                latex_from_string_list(
                    [latex_keyword(p, qc=True, link=False) for p in qc_parameters]
                ),
            ),
            (key_accuracy, latex_from_string(dictable[key_accuracy][0])),
            get_simple(dictable, key_remarks),
            get_simple(dictable, key_reqid),
            get_simple(dictable, key_parent_reqid),
        ]
    )
    lines_table = [f"{name}: & " + ls + "\\\\" for name, ls in dic3.items()]
    stable = "\n".join(lines_table)
    return f"""\\subsubsection{{{sname}}}
\\label{{recipe:{label_name_recipe}}}
\\label{{dataitem:{label_name_dataitem}}}

\\begin{{recipedef}}
{stable}
\\label{{tab:{label_table}}}
\\end{{recipedef}}
Table~\\ref{{tab:{label_table}}}
\\clearpage
"""


def write_recipe_calibration_plan(dataitem: Type[DataItem]):
    """Write recipe tex file to use for Calibration Plan.

    It is similar to the Recipe from the DRLD but has more rows.
    """
    fname = f"calibration_{dataitem.name_file_doc()}"
    # noinspection PyTypeChecker
    stex = get_tex_contents_recipe_calibration_plan(dataitem)
    fn_tex = PATH_DOCS_GENERATED / f"{fname}.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def write_datatitem_tex(dataitem: Type[DataItem]):
    """Write a dataitem to a tex file."""
    # TODO: Not all DataItems have a recipe attribute, perhaps refactor the
    #       class hierarchy to have a base class for processed data?

    fname = dataitem.name_file_doc()
    # noinspection PyTypeChecker
    stex_pro = get_tex_contents_dataitem_pro(dataitem)
    fn_rst = PATH_DOCS_GENERATED / f"{fname}.tex"
    with open(fn_rst, "w") as f:
        f.write(stex_pro)


def write_recipe_tex(dataitems):
    """Write tex files for Recipe."""
    # noinspection PyTypeChecker
    basedataitems = [di.get_base_dataitem() for di in dataitems]
    assert len(set(basedataitems)) == 1
    dataitem = basedataitems[0]
    stex_recipe_table = get_tex_contents_recipe(dataitem, dataitems)
    if stex_recipe_table:
        fname = dataitem.recipe.prop_type.name_eso()
        fn_rst = PATH_DOCS_GENERATED / f"{fname}.tex"
        with open(fn_rst, "w") as f:
            f.write(stex_recipe_table)


def write_dataitemraw_tex(raw, recipes: List[Type[Recipe]]):
    """Write a Raw dataitem to tex."""
    fname_raw = raw.name_file_doc()
    stex_raw = get_tex_contents_dataitem_raw(raw, recipes)
    fn_tex_raw = PATH_DOCS_GENERATED / f"{fname_raw}.tex"
    with open(fn_tex_raw, "w") as f:
        f.write(stex_raw)


def get_calibration_summary_tex():
    """Create calibration summary table."""
    # lines_placeholder = [
    #     # "    \\cct &\n",
    #     "    \\multicolumn{6}{|c|}{TODO: Placeholder} \\tabularnewline\n",
    #     "    \\hline\n",
    # ]
    # "\\rowcalibsummarysubheader{
    # In common for imaging and spectroscopy}\n",
    # "\\rowcalibsummarysubheader{
    # In common for standard and astrometric imaging}\n",
    rows = (
        [
            "% Rows copied from generated/calibration_plan_summary_imaging.tex\n",
            # get_tex_calibration_summary_row(Persistence),
            get_tex_calibration_summary_row(MasterDark, withaverageduration=True),
            # get_tex_calibration_summary_row(Nonlinearity),
            get_tex_calibration_summary_row(MasterFlat, withaverageduration=True),
            # "\\rowcalibsummarysubheader{Standard imaging only}\n",
            # get_tex_calibration_summary_row(IllumCorr),
            get_tex_calibration_summary_row(PhotomStdfield, withaverageduration=True),
            # get_tex_calibration_summary_row(Astrom),
            # "\\rowcalibsummarysubheader{Astrometric imaging only}\n",
        ]
        # + lines_placeholder
        + [
            # "\\rowcalibsummarysubheader{High Contrast imaging only}\n",
            get_tex_calibration_summary_row(Staroffset, withaverageduration=True),
            # get_tex_calibration_summary_row(StarcenterWaffle),
        ]
    )
    rows_static = [
        "% Rows copied from generated/calibration_plan_summary_imaging_static.tex\n",
        get_tex_calibration_summary_row(Nonlinearity, withaverageduration=False),
        get_tex_calibration_summary_row(IllumCorr, withaverageduration=False),
        get_tex_calibration_summary_row(DistortionWam, withaverageduration=False),
        get_tex_calibration_summary_row(DistortionElt, withaverageduration=False),
    ]
    # Assert to ensure no dataitem is forgotten.
    assert len(rows + rows_static) == len(CLASSES_CALIBRATION_PLAN) + 2
    rows_header = [
        "% Origin of columns in the digital design, e.g. for Dark / MasterDark:\n"
        "% Name: 'name' in masterdark.yaml, MasterDark.text_pro.name\n",
        "% OB Name: otype name in micado.vodsl, Dark.template.prop_type.name_eso()\n"
        "% Duration: 'duration' in masterdark.yaml if available, MasterdDark.text_pro.duration\n",
        "%            else 'duration' in dark.yaml, Dark.text_raw.duration\n",
        "% Raw data: otype name in micado.vodsl, Dark.name_eso()",
        "% Recipe: otype name in micado.vodsl, MasterDark.recipe.prop_type.name_eso()\n"
        "% Output data: otype name in micado.vodsl, MasterDark.name_eso()\n",
    ]
    return "".join(rows_header + rows), "".join(rows_header + rows_static)


def write_calibration_summary_tex():
    """Write calibration summary table."""
    stable, stable_static = get_calibration_summary_tex()

    fn_tex = PATH_DOCS_GENERATED / "calibration_plan_summary_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(stable)

    fn_tex = PATH_DOCS_GENERATED / "calibration_plan_summary_imaging_static.tex"
    with open(fn_tex, "w") as f:
        f.write(stable_static)


def write_dataitems_include_files(classes):
    """Write include files for dataitems and recipes and raws."""
    # TODO: Perhaps do a logical ordering instead of an alfabetical one.
    # Processed data.
    classes_pro = [
        dataitem
        for dataitem in classes
        # TODO: refactor this Depersisted hack.
        if not dataitem.name_file_doc().lower().endswith("base")
        or dataitem in [DepersistedBase]
    ]
    sincludes = "".join(
        sorted({f"\\input{{{dataitem.name_file_doc()}}}\n" for dataitem in classes_pro})
    )
    fn_tex = PATH_DOCS_GENERATED / "include_dataitems_imaging_pro.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)

    # Recipes.
    sincludes = "".join(
        sorted(
            {
                f"\\input{{{dataitem.recipe.prop_type.name_eso()}}}\n"
                for dataitem in classes
                if not dataitem.recipe.prop_type.name_eso().lower().endswith("base")
            }
        )
    )
    fn_tex = PATH_DOCS_GENERATED / "include_recipes_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)

    # Raw data.
    classes_raw = CLASSES_RAW
    classes_raw = [
        c
        for c in classes_raw
        if c not in [Raw, Sky] and not c.name_file_doc().lower().endswith("base")
    ]
    classes_ref = [AstromRef, PhotomRef, WamRef]
    sincludes = "".join(
        sorted(f"\\input{{{raw.name_file_doc()}}}\n" for raw in classes_raw)
    )
    fn_tex = PATH_DOCS_GENERATED / "include_dataitems_imaging_raw.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)

    classes_to_include_drl_structure_for = [
        di
        for di in CLASSES_INTERMEDIATE_DRL_FUNCTION
        if di not in classes_raw + classes_pro + classes_ref
    ]

    # DRL Data Structures.
    # Only include files for drl structures if the associated dataitems are
    # included as well.
    # Filename pattern from write_drlstructure_tex().
    fns_drl_tex = sorted(
        set(
            f"drldatastructure_{dataitem.name_eso_drl()}"
            for dataitem in classes_raw
            + classes_pro
            + classes_ref
            + classes_to_include_drl_structure_for
        )
    )
    sincludes = "".join(f"\\input{{generated/{fn}.tex}}\n" for fn in fns_drl_tex)
    fn_tex = PATH_DOCS_GENERATED / "include_drldatastructures_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)


def write_recipesteps_include_files(classes):
    """Write RecipeStep include files."""
    sincludes = "".join(sorted(f"\\input{{{c.name_file_doc()}}}\n" for c in classes))
    fn_tex = PATH_DOCS_GENERATED / "include_recipesteps_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)


def write_external_include_files(classes):
    """Write External include files."""
    sincludes = "".join(sorted(f"\\input{{{c.name_file_doc()}}}\n" for c in classes))
    fn_tex = PATH_DOCS_GENERATED / "include_external_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(sincludes)


def write_dataitems_tex(classes):
    """Loop over all DataItems and write tex files."""
    # Dictionaries of dictionaries, the latter to mimic an order-preserving set
    recipes_from_raw = defaultdict(lambda: {})
    dataitems_from_recipe = defaultdict(lambda: {})

    for dataitem_main in classes:
        write_datatitem_tex(dataitem_main)
        write_drlstructure_tex(dataitem_main)

        # Add the recipe of this processed dataitem to the list of recipes that
        # can process te associated raw.

        raw_main = dataitem_main.get_raw_class()
        recipe_main = dataitem_main.recipe.prop_type
        assert recipe_main != McdHciOptSciBase
        dataitems_from_recipe[recipe_main][dataitem_main] = True
        recipes_from_raw[raw_main][recipe_main] = True

        dataitems_derived = dataitem_main.get_derived_classes()
        for dataitem_derived in dataitems_derived:
            raw_derived = dataitem_derived.get_raw_class()
            recipe_derived = dataitem_derived.recipe.prop_type
            assert recipe_derived != McdHciOptSciBase
            dataitems_from_recipe[recipe_derived][dataitem_derived] = True
            recipes_from_raw[raw_derived][recipe_derived] = True

    for raw, recipes in recipes_from_raw.items():
        write_dataitemraw_tex(raw, list(recipes.keys()))
        if raw not in {Sky, Raw}:
            # Only concrete raws have to be described.
            write_drlstructure_tex(raw)

    for _, dataitems in dataitems_from_recipe.items():
        write_recipe_tex(list(dataitems.keys()))


def get_recipe_that_end_with_step(recipestep):
    """Get recipe that end with recipestep."""
    recipes_not_base = [
        recipe
        for recipe in RECIPES_ALL
        if not recipe.name_eso().endswith("base")
        # TODO: Fix McdImgDetrendHCI hack
        and recipe not in [McdImgDetrendHCI]
    ]
    recipes_end = [r for r in recipes_not_base if r.process.prop_type == recipestep]
    assert len(recipes_end) <= 1, recipes_end
    recipe = recipes_end[0] if recipes_end else None
    return recipe


def get_dataitem_produced_by_recipe(recipe):
    """Get the dataitem produced by recipe."""
    dataitems_with_recipe = [di for di in DATAITEMS_ALL if hasattr(di, "recipe")]
    dataitems = [di for di in dataitems_with_recipe if di.recipe.prop_type == recipe]
    if len(dataitems) == 0:
        dataitem = None
    elif len(dataitems) > 1:
        dataitem = dataitems[0]
        issuperclass = [issubclass(di, dataitem) for di in dataitems]
        assert all(issuperclass)
    else:
        dataitem = dataitems[0]
    return dataitem


def get_progenitor_from_recipestep(recipestep):
    """Get the progenitor of a recipestep that is not a step."""
    prev = recipestep.get_previous()
    prevstep = recipestep.get_previous_step()
    return None if prev == prevstep else prev


def latex_drlstructure(dataitem):
    """Get latex for drlstructure, with link."""
    if dataitem == Cat:
        return "cpl_table"
    if dataitem == Tab:
        return "cpl_propertylist"
    label_section = dataitem.name_eso().replace("_", "").replace("\\", "")
    return f"\\hyperref[drsstructure:{label_section}]{{{dataitem.name_eso_drl()}}}"


def patch_input_dataitem(dataitem):
    """Hack to get the right input dataitem for some special cases."""
    # TODO: Create general solution for this
    if dataitem in [
        DetrendFinalSiaiBase,
        DetrendFinalWaffle,
        DetrendFinalHCIBase,
        DetrendFinalPupil,
        DetrendFinalWaffle,
    ]:
        return DetrendFinal
    if dataitem in [DepersistedOpt, Raw]:
        return Img
    return dataitem


def get_output_dataitem_for_recipestep(recipestep):
    """Hack to get the output of the recipestep if it is not the last."""
    if issubclass(recipestep, Cat):
        return Cat
    if issubclass(recipestep, Tab):
        return Tab
    if issubclass(recipestep, Img):
        return Img

    raise ValueError(f"Cannot find output dataitem for {recipestep}")


def get_recipestep_tex(recipestep):
    """Get tex file contents for a RecipeStep."""
    fname = recipestep.name_file_doc()
    label_section = recipestep.name_eso().replace("_", "")
    name_section = latex_from_string(recipestep.name_eso())
    # TODO: is this needed: description = latex_from_string(recipestep.__doc__)
    parameters = [
        ParameterInfo.from_property(prop) for prop in recipestep.get_parameters()
    ]
    sparameters = (
        "\\newline ".join(
            latex_from_string(
                f"""
Name: \\PAR{{{param.name}}}
Type: {param.type}
Unit: {param.unit}
Default: {param.default}
Description: {param.description}
"""
            )
            for param in parameters
        )
        if parameters
        else "None"
    )

    qc = recipestep.get_qc_class()
    paramsqc = keywords_from_qc(qc) if qc else "None"
    sqc = (
        latex_from_string_list([latex_keyword(p, qc=True) for p in paramsqc])
        if qc
        else "None"
    )
    recipes = [
        recipe
        for recipe in RECIPES_ALL
        if recipestep in recipe.get_all_steps()
        and not recipe.name_eso().endswith("base")
    ]
    srecipes = latex_from_string_list([latex_recipe(r.name_eso()) for r in recipes])

    recipe_end = get_recipe_that_end_with_step(recipestep)
    dataitem_progenitor = get_progenitor_from_recipestep(recipestep)
    # TODO: Not all are Img
    dataitem_previous = Img if recipestep.get_previous_step() else None
    dataitem_input = dataitem_progenitor or dataitem_previous
    dataitem_input = patch_input_dataitem(dataitem_input)

    dataitems_attrs_input = recipestep.get_other_inputdata_full_attr(
        never_include_previous=True
    )

    dataitem_product = (
        get_dataitem_produced_by_recipe(recipe_end) if recipe_end else None
    )
    # TODO: Not all are Img
    dataitem_output = dataitem_product or get_output_dataitem_for_recipestep(recipestep)

    s_dataitem_input = f"const {latex_drlstructure(dataitem_input)} * progenitor"
    ss_dataitems_input = [
        f"const {latex_drlstructure(di)} * {attr}" for attr, di in dataitems_attrs_input
    ]
    s_dataitem_output = f"{latex_drlstructure(dataitem_output)} * product"
    s_input = "\n".join([s_dataitem_input] + ss_dataitems_input + [s_dataitem_output])

    # TODO: For special cases still do this?:
    # Other inputs: & \\texttt{{{latex_from_string(recipestep.text_recipestep.inputother)}}} \\\\

    s_qualityassessment = (
        "Through QC parameters"
        if recipestep.text_recipestep.qualityassessment == "None"
        else recipestep.text_recipestep.qualityassessment
    )
    s_errorconditions = (
        "See \cite{drlvt}."
        if recipestep.text_recipestep.errorconditions == "None"
        else recipestep.text_recipestep.errorconditions
    )
    s_unittests = (
        "See \cite{drlvt}."
        if recipestep.text_recipestep.unittests == "None"
        else recipestep.text_recipestep.unittests
    )

    sfigure = f"""
\\input{{generated/{fname}_tikz.tex}}

% \\begin{{figure}}[ht]
%     \\centering
%     \\includegraphics[scale=0.5]{{generated/{fname}}}
%     \\caption{{{name_section}}}
%     \\label{{fig:{fname}}}
% \\end{{figure}}
    """
    # TODO: Decide what todo with Description: & {description} \\\\
    # TODO: Quality assessment, error conditions, unit tests
    return f"""\\clearpage
\\subsubsection{{{name_section}}}\\label{{dataitem:{label_section}}}
{sfigure}
\\begin{{recipedef}}
Name: & {name_section} \\\\
Purpose: & {latex_from_string(recipestep.text_recipestep.purpose)} \\\\
Used in recipes: & {srecipes} \\\\
Working remarks: & {latex_from_string(recipestep.text_recipestep.workingremarks)} \\\\
Function Parameters: & {sparameters} \\\\
%Input FITS files: & {latex_from_string(recipestep.text_recipestep.inputfits)} \\\\
Inputs: & \\texttt{{{latex_from_string(s_input)}}} \\\\
%QC outputs: & {sqc} \\\\
%Output FITS files: & {latex_from_string(recipestep.text_recipestep.outputfits)} \\\\
Other outputs: & \\texttt{{{latex_from_string(recipestep.text_recipestep.outputother)}}} \\\\
General description: & {latex_from_string(recipestep.text_recipestep.generaldescription)} \\\\
Mathematical description: & {latex_from_string(recipestep.text_recipestep.mathematicaldescription)} \\\\
Quality assessment: & {latex_from_string(s_qualityassessment)} \\\\
Error conditions: & {latex_from_string(s_errorconditions)} \\\\
Unit tests: & {latex_from_string(s_unittests)} \\\\
\\end{{recipedef}}
    """


def write_recipestep_tex(recipestep):
    """Write a tex file for a RecipeStep."""
    stex = get_recipestep_tex(recipestep)
    fname = recipestep.name_file_doc()
    fn_tex = PATH_DOCS_GENERATED / f"{fname}.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def recipesteps_from_dataitems(dataitems):
    """Get RecipeStep classes from dataitem classes."""
    recipesteps = []
    for dataitem in dataitems:
        recipestep = dataitem.recipe.prop_type.process.prop_type
        while recipestep is not None:
            if recipestep not in recipesteps:
                recipesteps.append(recipestep)
            recipestep = recipestep.get_previous_step()
    return recipesteps


# TODO: Refactor keywords_forbidden.
keywords_forbidden = {
    "XTENSION",
    "BITPIX",
    "NAXIS",
    "FILENAME",
    "NAXIS1",
    "NAXIS2",
    "EXTNAME",
    "INHERIT",
    "SIMPLE",
}


def tex_keywords(kws):
    """Fix keywords."""
    kws2 = [
        latex_keyword(kw.replace("HIERARCH ESO ", "").replace(" ", "."))
        for kw in kws
        if " " not in kw and kw not in keywords_forbidden
    ]
    return (" (" + " ".join(kws2) + ")") if kws2 else ""


def get_drlstructure_tex(dataitem: Type[DataItem]):
    """Get DRL data structure tex."""
    name_section = latex_from_string(dataitem.name_eso_drl())
    label_section = dataitem.name_eso().replace("_", "")
    tex_dataitem = latex_dataitem(latex_from_string(dataitem.name_eso()))
    layer_names, layer_names_opt = dataitem.layer_names()

    # Sanity check to ensure that other combinations of layer names are
    # properly supported once they are added.
    assert not layer_names_opt
    sln_pro = {KEY_SCI, KEY_DQ, KEY_ERR}
    sln_raw = {KEY_SCI}
    sln_cat = {KEY_CAT}
    sln_tab = {KEY_TAB}
    sln_nonlin = {KEY_DATAN, KEY_CHI2, KEY_DOF}
    sln = set(layer_names)
    assert sln in (sln_pro, sln_raw, sln_cat, sln_tab, sln_nonlin)
    assert KEY_CNT not in layer_names
    assert KEY_PERS not in layer_names
    assert KEY_BG not in layer_names

    keywords_primary = tex_keywords(dataitem.fits_keywords_c())
    if hasattr(dataitem, "extensions"):
        keywords_extensions = tex_keywords(
            dataitem.extensions.prop_type.fits_keywords_c()
        )
    else:
        keywords_extensions = ""

    lmin, lmax = dataitem.layers_number()
    one_image = lmin == 1 == lmax
    fields_opt = [
        f"cpl_propertylist * keywords: Primary keywords{keywords_primary}",
        "hdrl_image * image: Image"
        if one_image and sln in (sln_pro, sln_raw)
        else None,
        "hdrl_imagelist * images: Images"
        if not one_image and sln in (sln_pro, sln_raw)
        else None,
        f"cpl_propertylist * extkeywords: Extension keywords{keywords_extensions}"
        if one_image
        else None,
        f"cpl_propertylist ** extkeywords: Extension keywords{keywords_extensions}"
        if not one_image
        else None,
        # All standard image tables are kept in the headers / propertylists.
        # TODO: Check with HCI and AI.
        # "cpl_table * table: Table" if one_image and sln in (sln_tab,)
        # else None,
        # "cpl_table ** tables: Tables" if not one_image and sln in (sln_tab,)
        # else None,
        "cpl_table * catalog: Catalog" if one_image and sln in (sln_cat,) else None,
        "cpl_table ** catalogs: Catalogs"
        if not one_image and sln in (sln_cat,)
        else None,
    ]
    fields_tex = [
        f"    \\item \\texttt{{{latex_from_string(s)}}}" for s in fields_opt if s
    ]
    fields_flat_tex = "\n".join(fields_tex)
    return f"""\\subsubsection{{{name_section}}}\\label{{drsstructure:{label_section}}}
This structure represents all contents of a {tex_dataitem}.
The fields are:
\\begin{{enumerate}}
{fields_flat_tex}
\\end{{enumerate}}
"""


def write_drlstructure_tex(dataitem: Type[DataItem]):
    """Write DRL data structure to file."""
    stex = get_drlstructure_tex(dataitem)
    fname = f"drldatastructure_{dataitem.name_eso_drl()}"
    fn_tex = PATH_DOCS_GENERATED / f"{fname}.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def latex_from_template(template, dataitem, include_headers=False):
    """Get latex string from template."""
    kws_prim = sorted(dataitem.fits_keywords_c())
    kws_prim = [k for k in kws_prim if "FILENAME" not in k]
    sts_pkws = [f"""  \\item {latex_keyword(name)}\n""" for name in kws_prim]
    st_primary = (
        f"""Primary header keywords of main data item produced by template:
\\begin{{itemize}}
    \\setlength{{\\itemsep}}{{0pt}}
    \\setlength{{\\parskip}}{{0pt}}
{''.join(sts_pkws)}
\\end{{itemize}}
"""
        if kws_prim
        else ""
    )

    st_extension = ""
    if hasattr(dataitem, "extensions"):
        kws_ext = sorted(dataitem.extensions.prop_type.fits_keywords_c())
        kws_ext = [k for k in kws_ext if "FILENAME" not in k]
        sts_ekws = [f"""  \\item {latex_keyword(name)}\n""" for name in kws_ext]
        st_extension = (
            f"""Extension header keywords of main data item produced by template:
\\begin{{itemize}}
    \\setlength{{\\itemsep}}{{0pt}}
    \\setlength{{\\parskip}}{{0pt}}
{''.join(sts_ekws)}
\\end{{itemize}}
    """
            if kws_ext
            else ""
        )

    if not include_headers:
        st_primary = ""
        st_extension = ""

    parameters = latex_from_string_list(
        [
            f"\\PAR{{{p}}}" if p != "N/A" else "N/A"
            for p in parameters_from_template(template).split(", ")
        ]
    )

    return """\\subsubsection{{{name}}}\\label{{{sectionlabel}}}
\\begin{{recipedef}}
Name & {name} \\\\
Description & {description} \\\\
\\FITS{{CATG}} & \\FITS{{{catg}}} \\\\
\\FITS{{TYPE}} & \\FITS{{{type}}} \\\\
\\FITS{{TECH}} & \\FITS{{{tech}}} \\\\
Parameters & {parameters} \\\\
Main Dataitem & {dataitem} \\\\
\\end{{recipedef}}

{st_primary}

{st_extension}
\\clearpage
""".format(
        name=latex_template(template.name_eso()),
        sectionlabel="template:"
        + template.name_eso()
        .replace("_", "")
        .replace("\\", "")
        .replace("[", "")
        .replace("]", ""),
        description=latex_from_string(template.description_oneline()),
        catg=template.dpr__catg.default,
        type=template.dpr__type.default,
        tech=template.dpr__tech.default,
        parameters=parameters,
        dataitem=latex_dataitem(latex_from_string(dataitem.name_eso())),
        st_primary=st_primary,
        st_extension=st_extension,
    )


def get_template_table():
    """Get template table from Python.

    Create a reference for all templates to here, because there is no separate
    section per template at the moment.
    """
    templatesplus = get_template_classes_with_raw()
    templatesplus.sort(key=lambda t: t[0].name_eso())
    rows = [
        (
            r"        {name} & {dataitem} & "
            r"\FITS{{{catg}}} & \FITS{{{type}}} & \FITS{{{tech}}} & "
            r"{recipe} & {successor}"
            r" \tabularnewline"
        ).format(
            name=latex_template(latex_from_string(template.name_eso())),
            dataitem=latex_dataitem(latex_from_string(raw.name_eso())),
            catg=template.dpr__catg.default,
            type=template.dpr__type.default,
            tech=template.dpr__tech.default,
            recipe=latex_recipe(latex_from_string(recipe.name_eso())),
            successor=latex_dataitem(latex_from_string(successor.name_eso())),
        )
        for template, raw, recipe, successor in templatesplus
        # TODO: Decide what to do with these virtual templates.
        if "virtual" not in template.name_eso()
    ]
    # For some unknown reason the \input only works if the last hline is within
    # the included file.
    # TODO: Remove MICINSNAME hack from here.
    return ("\n        \\hline\n".join(rows) + "    \\hline\n").replace(
        r"MICINSNAME\_", r"MICADO\_"
    )


def get_template_parameters_table():
    """Get table with template parameters."""
    templatesplus = get_template_classes_with_raw()
    templatesplus.sort(key=lambda t: t[0].name_eso())


def get_templates_appendix(
    include_summary_table=False,
    include_explanation=False,
):
    """Get appendix for templates."""
    templatesplus = get_template_classes_with_raw()
    templatesplus.sort(key=lambda t: t[0].name_eso())
    # s_table = get_template_table()
    s_table = r"\input{template_summary_imaging_rows}"
    s_table_full = (
        r"""\begin{sidewaystable}[ht!]
    \small
    \centering
    \begin{tabular}{|l|l|l|l|l|l|l|}
        \hline
        \rowcolor{GreenYellow} \textbf{\FITS{Name}} & \textbf{\PROD{Raw}} &
        \textbf{\FITS{CATG}} & \textbf{\FITS{TYPE}} & \textbf{\FITS{TECH}} &
        \textbf{\REC{CPL Recipe}} & \textbf{\PROD{Processed}}\tabularnewline
        \hline
"""
        + s_table
        + r"""
    \end{tabular}
    \caption{
        Imaging templates with their main DataItem and \textbf{DPR} keywords.
        Only data that is needed for the \ac{DFS} is listed here.
    }
    \label{tab:templates_imaging}
\end{sidewaystable}
"""
    )

    s_explanation = r"""

All FITS header keyword of the main data item are listed below.
The FITS header keywords contain all the information necessary to process the data as well as simulate it.

The following hierarchies are used in the hierarchical keywords below:
\begin{itemize}
    \setlength{\itemsep}{0pt}
    \setlength{\parskip}{0pt}
    \item \FITS{ATM}: Atmosphere
    \item \FITS{DET}: DetectorArray (in primary) or Detector (in extension or within \FITS{DET})
    \item \FITS{DETi}: Detector in a DetectorArray
    \item \FITS{DAR}: Dark Current
    \item \FITS{SHO}: Shotnoise (simulator only)
    \item \FITS{DPR}: Catagory Data Product
    \item \FITS{INS}: Instrument
    \item \FITS{FIL}: Filter
    \item \FITS{MOD}: Mode Mirrors
    \item \FITS{OBS}: Observation
    \item \FITS{REL}: Relay Optics
    \item \FITS{STA}: Static Mirrors
    \item \FITS{SUR}: Surface (e.g. within \FITS{MOD} or \FITS{STA})
    \item \FITS{TEL}: Telescope
    \item \FITS{TPL}: Template
    \item \FITS{VIB}: Residual vibration (of telescope)
\end{itemize}
"""

    s_templates = [
        latex_from_template(template, raw)
        for template, raw, recipe, successor in templatesplus
    ]
    s_templates_joined = "\n\n".join(s_templates)
    return (
        r"""\subsection{Imaging Templates}\label{sec:imagingtemplates}

"""
        + (s_table_full if include_summary_table else "")
        + (s_explanation if include_explanation else "")
        + s_templates_joined
    )
    # TODO: Automate the table above


def get_processed_and_recipe_from_raw(dataitem: Type[DataItem]):
    """Get the main processed data item and the recipe from this dataitem."""
    dataitems_all = [
        v.aclass
        for v in classcache.values()
        if issubclass(v.aclass, DataItem)
        # TODO: Do something sensible for Nonlinearity and online processing
        and v.aclass not in [Nonlinearity, DpsicsStaroffsetSci]
    ]
    successors = [v for v in dataitems_all if v.get_progenitor_class() == dataitem]
    assert len(successors) == 1
    successor = successors[0]
    recipe = successor.recipe.prop_type
    return recipe, successor


def get_template_classes_with_raw():
    """Get template classes with associated raw classes."""
    # Get all template classes from src/micado/templates.
    template_classes1: List[Template] = get_template_classes()
    # TODO: Perhaps sort the template_classes? But in what order?
    #       - Alphabetical?
    #       - First ack, then cal, then sci?
    #       - In order as they appear in association matrices?
    classes_raw_ics = CLASSES_RAW
    # TODO: Find better way to include/exclude DPSICS
    # CLASSES_ICS is not included in classes_raw_ics anymore.
    template_classes = [
        temp for temp in template_classes1 if temp not in [MicadoCoroAcqStar]
    ]

    # Get all templates from the raw data classes.
    templates_raw = [raw.template.prop_type for raw in classes_raw_ics]
    # Sanity check to ensure that these are the same.
    templates_extra_classes = set(template_classes) - set(templates_raw)
    templates_extra_raw = set(templates_raw) - set(template_classes)
    assert not templates_extra_raw, templates_extra_raw
    assert not templates_extra_classes, templates_extra_classes

    templatesplus = [
        (
            raw.template.prop_type,
            raw,
            get_processed_and_recipe_from_raw(raw)[0],
            get_processed_and_recipe_from_raw(raw)[1],
        )
        for raw in classes_raw_ics
    ]

    return templatesplus


def get_keywords_labels(name):
    """Get labels for keyword."""
    # Add labels like coeff00 and coeff7
    labels_extra_xx = (
        "".join(
            f"""\\label{{parameter:{name.replace("i", "%02i" % i).lower().replace(" ", ".").replace("_", "")}}}"""
            for i in range(100)
        )
        if name.endswith("i")
        else ""
    )
    labels_extra_x = (
        "".join(
            f"""\\label{{parameter:{name.replace("i", "%i" % i).lower().replace(" ", ".").replace("_", "")}}}"""
            for i in range(10)
        )
        if name.endswith("i")
        else ""
    )
    # Replace e.g. TPLi START with TPL START
    label_extra_noi = f"""\\label{{parameter:{name.replace("i", "").lower().replace(" ", ".").replace("_", "")}}}"""
    # TODO: FIX PV HACK
    labels_extra_pv = (
        r"\label{parameter:pv1i}\label{parameter:pv2i}\label{parameter:pv1}"
        r"\label{parameter:pv2}\label{parameter:pv11}\label{parameter:pv22}"
        if name.startswith("PV")
        else ""
    )
    # All TPL keywords are repeated as OBS TPLi keywords, but no need to repeat
    # them in the documentation.
    labels_extra_tpl = f"""\\label{{parameter:{
        name.replace("TPL ", "OBS TPL ").lower().replace(" ", ".").replace("_", "")}}}"""
    label_main = (
        f"""\\label{{parameter:{name.lower().replace(" ", ".").replace("_", "")}}}"""
    )
    labels_extra_filt = (
        r"\label{parameter:ins.filti.name}"
        r"\label{parameter:ins.filt1.name}"
        r"\label{parameter:ins.filt2.name}"
        if "ins.filt.name" in label_main
        else ""
    )
    labels = (
        label_main
        + labels_extra_x
        + labels_extra_xx
        + labels_extra_pv
        + labels_extra_filt
    )
    # label_extra_noi can cause a duplicate label
    for lab in [labels_extra_tpl, label_extra_noi]:
        if lab not in labels:
            labels += lab
    return labels


def get_fits_keywords_one_tex(di):
    """Get fits keywords for one dataitem."""
    di_base = di.base_doc_kw()
    st_base = (
        f"""All headers from \\hyperref[fitskeywords:{di_base.name_eso(includebase=True).replace("_", "")}]"""
        + f"""{{{latex_from_string(di_base.name_eso(includebase=True))}}}."""
        if di_base
        else ""
    )
    kws_prim, kws_ext = di.fits_keywords_new_c()
    kws_prim = [k for k in kws_prim if "FILENAME" not in k]
    kws_ext = [k for k in kws_ext if "FILENAME" not in k]

    sts_pkws = [f"""  \\item {latex_keyword(name)}\n""" for name in kws_prim]
    st_primary = (
        f"""Primary:
    \\begin{{itemize}}
    \\setlength{{\\itemsep}}{{0pt}}
    \\setlength{{\\parskip}}{{0pt}}
{''.join(sts_pkws)}
    \\end{{itemize}}
"""
        if kws_prim
        else ""
    )

    sts_ekws = [f"""  \\item {latex_keyword(name)}\n""" for name in kws_ext]
    st_extension = (
        f"""Extensions:
    \\begin{{itemize}}
    \\setlength{{\\itemsep}}{{0pt}}
    \\setlength{{\\parskip}}{{0pt}}
{''.join(sts_ekws)}
    \\end{{itemize}}
"""
        if kws_ext
        else ""
    )

    return (
        f"""\\subsubsection{{{latex_from_string(di.name_eso(includebase=True))}}}"""
        f"""\\label{{fitskeywords:{di.name_eso(includebase=True).replace("_", "")}}}

{st_base}

{st_primary}

{st_extension}
"""
    )


def get_fits_keywords_tex(classes):
    """Get fits keywords for classes."""

    def count_base_depth(di):
        name = di.name_eso(includebase=True)
        d = 0
        while di := di.base_doc_kw():
            d += 1
        return d, name

    # Add some base classes as they are used in the header hierarchy.
    # Also add raw classes.
    classes2 = (
        [DataItem, StarcenterBase, Opt, SkyHCIOnline, AstromRef, PhotomRef, WamRef]
        + classes
        + [c.get_raw_class() for c in classes]
    )
    # Filter out the None-raw classes and duplicates.
    # Order by how deep the DataItems are in the header hierarchy.
    classes_sorted = sorted(
        set(c for c in classes2 if c is not None), key=count_base_depth
    )
    return "\n\n".join(get_fits_keywords_one_tex(di) for di in classes_sorted)


def write_fits_keywords_tex(classes):
    """Write TeX file with keywords per dataitem."""
    stex = get_fits_keywords_tex(classes)
    fn_tex = PATH_DOCS_GENERATED / "keywords_list_dataitems_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def get_esokeywords_tex(qc=False):
    """Get TeX version of ESO keywords.

    No QC if qc==False, only QC if qc==True.
    """
    # Check which ones are used.
    paths_tex = PATH_DOCS_GENERATED.glob("*.tex")
    kws_used = [
        k
        for p in paths_tex
        if p.name
        not in [
            "keywords_list_dataitems_imaging.tex",
            "template_summary_imaging.tex",
        ]
        for k in re.findall(r"\\hyperref\[parameter:(.*?)]", p.open().read())
    ]

    return [
        f"""\\subsubsection{{{latex_from_string(name.replace(" ", "."))}}}{get_keywords_labels(name)}
\\begin{{recipedef}}
Name & {latex_from_string(kwd['name'])} \\tabularnewline
Class & {latex_from_string(kwd['class'])} \\tabularnewline
Context & {latex_from_string(kwd['context'])} \\tabularnewline
Type & {latex_from_string(kwd['type'])} \\tabularnewline
Value & {latex_from_string(kwd['value'])} \\tabularnewline
Unit & {latex_from_string(kwd['unit'])} \\tabularnewline
Comment & {latex_from_string(kwd['comment'])} \\tabularnewline
Default & {latex_from_string(kwd['default'])} \\tabularnewline
Range & {latex_from_string(kwd['range'])} \\tabularnewline
Description & {latex_from_string(kwd['description'])} \\tabularnewline
\\end{{recipedef}}
"""
        for name, kwd in esokeywords.items()
        if (qc and name.startswith("QC "))
        or (
            not qc
            and not name.startswith("QC ")
            and any(
                k in kws_used
                for k in re.findall(r"{parameter:(.*?)}", get_keywords_labels(name))
            )
        )
    ]


def get_recipeqcparams_tex():
    """Create table with Recipe name vs QC parameters."""
    qc_from_recipe = {}
    for dataitem in CLASSES_DATAITEMS:
        if not hasattr(dataitem, "recipe"):
            continue
        recipe = dataitem.recipe.prop_type
        if recipe not in qc_from_recipe:
            qc_from_recipe[recipe] = []
        qc = dataitem.get_qc_class()
        if not qc:
            continue
        for qcparam in keywords_from_qc(qc):
            if qcparam not in qc_from_recipe[recipe]:
                qc_from_recipe[recipe].append(qcparam)

    s_rows = []
    for recipe, qcparams in qc_from_recipe.items():
        qcparams = qcparams or [""]
        s_rows.append(r"    \hline")
        s_rows.append(
            rf"    {latex_recipe(recipe.name_eso(), link=False)} & {latex_keyword(qcparams[0], link=False, qc=True)} & \\"
        )
        for qcparam in qcparams[1:]:
            s_rows.append(rf"     & {latex_keyword(qcparam, link=False, qc=True)} & \\")
    s_rows_joined = "\n".join(s_rows)
    s_table = fr"""
\begin{{longtable}}{{|l|l|l|}}
    \hline
    \textbf{{Recipe}} & \textbf{{QC parameter}} & Info \\
{s_rows_joined}
    \hline
    \caption{{Recipes with the QC parameters they produce.}}
    \label{{tab:recipeswithqc}}
\end{{longtable}}
"""
    return s_table


def write_recipeqcparams_tex():
    """Write Recipe-QC params table."""
    stex = get_recipeqcparams_tex()
    fn_tex = PATH_DOCS_GENERATED / "drlvt_recipe_vs_qc_params.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def write_esokeywords_tex():
    """Write TeX version of ESO keywords."""
    stex = "\n\n".join(get_esokeywords_tex())
    fn_tex = PATH_DOCS_GENERATED / "keywords_list_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)

    stex = "\n\n".join(get_esokeywords_tex(qc=True))
    fn_tex = PATH_DOCS_GENERATED / "keywords_list_qc.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def write_tex_essential_raw_keywords():
    """Write TeX version of essential RAW keywords list."""
    s_tex = get_tex_essential_raw_keywords()
    fn_tex = PATH_DOCS_GENERATED / "keywords_list_raw_essential.tex"
    with open(fn_tex, "w") as f:
        f.write(s_tex)


def get_recipe_decomposition_tex():
    """Get decomposition of recipes into drl functions."""
    rows = []
    for recipe in RECIPES_ALL:

        if "base" in recipe.__name__.lower():
            continue

        # TODO: Find better way to include/exclude DPSICS
        if issubclass(recipe, DpsicsStaroffsetBase.recipe.prop_type):
            continue

        if rows:
            rows.append("    \\hline")

        name_recipe = latex_recipe(recipe.name_eso(), include_normal_reference=True)
        recipesteps = recipe.get_all_steps()
        # Ensure this recipe has steps.
        assert recipesteps
        names_steps = [
            latex_recipestep(rs.name_eso(), include_normal_reference=True)
            for rs in recipesteps
        ]
        rows.append(f"    {name_recipe} & {names_steps[0]} \\tabularnewline")
        for name_step in names_steps[1:]:
            rows.append(f"      & {name_step} \\tabularnewline")

    return "\n".join(rows)


def write_recipe_decomposition_tex():
    """Write TeX version of recipe decomposition."""
    stex = get_recipe_decomposition_tex()
    fn_tex = PATH_DOCS_GENERATED / "cplrecipe_to_drlfunction.tex"
    with open(fn_tex, "w") as f:
        f.write(stex)


def main():
    """Write dataitems."""
    # TODO: Find better way to deal with DepersistedBase
    dataitems_img = [
        di
        for di in CLASSES_DATAITEMS
        if not issubclass(di, DepersistedBase) or di == DepersistedBase
    ]
    write_dataitems_tex(dataitems_img)

    write_dataitems_include_files(dataitems_img)

    recipesteps = recipesteps_from_dataitems(dataitems_img)
    for recipestep in recipesteps:
        write_recipestep_tex(recipestep)

    write_recipesteps_include_files(recipesteps)

    for external in CLASSES_EXTERNAL + CLASSES_INTERMEDIATE_DRL_FUNCTION:
        write_datatitem_external_tex(external)
        write_drlstructure_tex(external)
    write_external_include_files(CLASSES_EXTERNAL)

    # TODO: Write keywords table
    # skeywords = get_keywords_table()
    # fn_rst_keywords = PATH_DOCS_GENERATED / "keywordstable.tex"
    # with open(fn_rst_keywords, 'w') as f:
    #    f.write(skeywords)

    for dataitem in CLASSES_CALIBRATION_PLAN:
        write_recipe_calibration_plan(dataitem)

    write_calibration_summary_tex()

    s_template_appendix = get_templates_appendix()
    fn_tex = PATH_DOCS_GENERATED / "template_summary_imaging.tex"
    with open(fn_tex, "w") as f:
        f.write(s_template_appendix)

    s_template_appendix_rows = get_template_table()
    fn_tex = PATH_DOCS_GENERATED / "template_summary_imaging_rows.tex"
    with open(fn_tex, "w") as f:
        f.write(s_template_appendix_rows)

    write_tex_essential_raw_keywords()

    write_recipeqcparams_tex()

    write_recipe_decomposition_tex()

    write_fits_keywords_tex(dataitems_img)

    write_esokeywords_tex()


if __name__ == "__main__":
    main()
